from labton.labton_backend.data_handler import DatabaseHandler
from labton.labton_backend.helper_funcs import update_history, fecth_new_paragraph_id
from flask import session, render_template, request, redirect, url_for, jsonify, send_from_directory, abort

class ActionHandler(DatabaseHandler):
    def __init__(self):
        super().__init__()
        
    def action_flow_annotation(self):
        if 'username' in session:
                print("_"*70)
                
                if request.method == 'POST':
                    print("*"*70)
                    post = request.get_json()
                    print("POST:")
                    print(post, end="\n")
                    if post["action"] == "annotate":
                        values = [post["annotation"],session["username"], 
                                post["annotering_text"].strip() if post["annotering_text"] else None,
                                post["paragraph_id"]]
                        self.add_annotation(values, "annotations")
                        print(session)

                    elif post["action"] == "review":
                        print("||"*70)
                        print("THIS IS REVIEW")
                        #https://stackoverflow.com/questions/15473626/make-a-post-request-while-redirecting-in-flask/15480983#15480983
                        print(url_for('review'))
                        return redirect(url_for('review'))

                    elif post["action"] == "logout":
                        #Save and close user_session table
                        session.clear()
                        return redirect(url_for('login'))

                    elif post["action"] in ["back", "begining", "forward", "end"]:
                        print("BUTTON PRESSED")
                        paragraph_id = fecth_new_paragraph_id(session, post, post["action"])
                        history_dive = True

                    else:
                        pass

                #Hvis session stadig er aktiv og der  ikke er er defieneret et paragraph_id 
                # (dette sker hvis man har lukket browser og er logget ind igen)
                #Så hent sidste entry i annotation_history som paragraph_id
                try: paragraph_id
                except NameError: 
                    if "history_dive" in session: 
                        paragraph_id = session["annotation_history"][-1]
                    else: pass
                
                dive_check = session["history_dive"] if "history_dive" in session else False
                print(f"ARE WE DIVING?? {dive_check}")
                #Hvis man er på den nyeste sætning, så er man ude af historie dykket
                #OBS! herved vil man aldrig lande på den sidste sætning igen :/ må løses
                if "history_dive" in session and session["history_dive"] is True:
                    print("GOING TO THE PAST")
                    
                    paragraph_text, paragraph_id, document_name, page_nr = self.fetch_from_db(["paragraph_text", 
                                                    "paragraph_id", "document_name", "page_nr"], "annotations", 
                                                    f"paragraph_id = {paragraph_id}" , one=True)
                    if session["annotation_history"][-1] == paragraph_id:
                        session["history_dive"] = False
                    print(f"ARE WE STILL DIVING?? {session['history_dive']}")
                
                else:
                    print("STAYING IN THE FUTURE")                        
                    paragraph_text, paragraph_id, document_name, page_nr = self.fetch_from_db(["paragraph_text", 
                                                                    "paragraph_id", "document_name", "page_nr"], 
                                                                    "annotations", "correct_annotation IS null", 
                                                                    one=True)
                print("SENT SUPPOSED TO BE ON SCREEN")
                print(paragraph_text)
                update_history(session, paragraph_id)
                return render_template("annotering.html", 
                                        session=session,
                                        sent=paragraph_text,
                                        paragraph_id=paragraph_id,
                                        png_name= f"{document_name}000{page_nr}-{page_nr}.png",
                                        classes=self.config["classes"])
        else:
            return redirect(url_for('login'))

    def action_flow_login(self):
        if request.method == 'POST':
            username = request.form['username']
            password = request.form['password']
            #TODO: implement login here
            if username: #authenticate(str(username), str(password))
                session['username'] = request.form['username']
                return redirect(url_for('annotering'))
            else:
                return """
                    <head>
                    <style type="text/css">
                        body {
                        text-align: center;
                        background-size: 100%;
                        }
                        h1, h2 {
                        color: black;
                        font-family: "Courier New";
                        }
                    </style>
                    </head>
                    <h2>You need to type in a name to annotate as<h2>
                    <form action="/login">
                            <input type="submit" value="Try again" />
                    </form> 
                """
                'Invalid username/password'
        else:
            return render_template("login.html")
    
    def action_logout(self):
        # remove the username from the session if it's there
        session.pop('username', None)
        return redirect(url_for('annotering'))
    
    def action_flow_review(self):
        if 'username' in session:
            print("GOING TO REVIEW")
            if request.method == 'POST':
                post = request.get_json()
                print("Review POST:")
                print(post, end="\n")

                if post["action"] == "submit":
                    paragraph_ids = post["paragraph_ids"] 
                    who_verified = post["who_verified"]
                    verifications = post["verifications"]
                    for values in zip(who_verified, verifications, paragraph_ids):
                        self.verify_annotation(values, "annotations")
                        print(values)
            paragraph_texts = self.fetch_from_db(["paragraph_id", "paragraph_text", "correct_annotation", "extract"], "annotations",
                                        "correct_annotation NOT NULL AND who_verified IS NULL " +\
                                        f"AND who_annotated <> '{session['username']}'", one=False)
            #OBS! DataHandler.fetch_from_db returns (None, None, None, None) when nothing found in DB
            #We avoid empty lines in the reviewpage with the following line of code
            paragraph_texts =  [i for i in paragraph_texts if i]
            return render_template("review.html", enumerated_paragraph_texts = enumerate(paragraph_texts), 
                                                        total=len(paragraph_texts), session=session)
        else:
            return redirect(url_for('login'))


from labton.labton_backend.connection_handler import SQliteConnection
import pandas as pd


class DatabaseHandler(SQliteConnection):
    def __init__(self):
        super().__init__()
        
    def create_database(self, data_source, csv_sep, is_df=False):
        if is_df:
            df_text = data_source
        else:
            df_text = pd.read_csv(data_source, 
                                sep=csv_sep, keep_default_na=False,
                                engine='python', skipinitialspace=True,
                                names=["paragraph_text", "document_name", "page_nr"])
        df_text["paragraph_id"] = df_text.index
        df_text.to_sql(name = "annotations", con = self.conn)
        columns_to_add = [
                            "correct_annotation CHAR(50)",
                            "who_annotated CHAR(50)",
                            "extract VARCHAR(10)",
                            "who_verified CHAR(50)",
                            "verification CHAR(50)"
                        ]
        
        add_col = lambda col:  self.conn.execute(f"ALTER TABLE annotations ADD {col}")
        for col in columns_to_add: add_col(col)     
    
    def fetch_from_db(self, columns, table, conditions=None, one=False):
        #Conditions must be string
        if type(columns) == str: columns = [columns]
        #https://docs.python.org/2/library/sqlite3.html#sqlite3.Connection.row_factory
        conditions_sql = f"WHERE {conditions}" if conditions else ""
        self.curr.execute(f"""
            SELECT {", ".join(columns)} 
            FROM {table}
            {conditions_sql};
        """)
        out = self.curr.fetchall() if not one else self.curr.fetchone()
        return(out) if out else (None, None, None, None)
    
    def add_annotation(self, values, table):
        #OBS! make naming convention of databasecolumns in english
        self.conn.execute(f"""
        UPDATE {table} SET 
                    correct_annotation = ?, 
                    who_annotated = ?, 
                    extract = ?
        WHERE paragraph_id == ?;""", values)
    
    def verify_annotation(self, values, table):
        self.conn.execute(f"""
        UPDATE {table} SET 
                    who_verified = ?,
                    verification = ? 
        WHERE paragraph_id == ?;
        """, values)

    def load_annotations_to_pandas(self):
        return(pd.read_sql("SELECT * FROM annotations", self.conn))


from functools import wraps
from flask import Flask
from pandas import DataFrame
from labton.labton_backend.connection_handler import SQliteConnection
from labton.labton_backend.config_file_handler import ConfigHandler
from labton.labton_backend.data_handler import DatabaseHandler

def with_sqlite_conn(f):
    """
    Wrapper to wrap around function that executes 
    SQLite code on the database whos path is defined 
    in the config variable path_db_file

    Args:
        f (function): function that excecutes to SQL-db
    Returns:
        function: function wrapped such that 
                open/close/commit is handeled automatically
    """
    @wraps(f)
    def wrapper(*args, **kwds):
        with SQliteConnection() as conn_obj:
            return f(*args, conn=conn_obj.conn, curr=conn_obj.curr, **kwds)
    return wrapper

def update_history(session, paragraph_id):
    #Hvis det ikke er første annotering i sessionen
    if "annotation_history" in session:
        history = session["annotation_history"]
        current_id = paragraph_id #post["paragraph_id"]
        #Hvis annoterede sætning allerede er blevet annoteret
        if current_id not in history:
            history += [current_id]
            session["annotation_history"] = history
        else:
            pass
            # idx = history.index(current_id) 
            # del history[idx]

    else: 
        session["annotation_history"] = []
    print(session["annotation_history"])
    return session

def fecth_new_paragraph_id(session, post, move):
    # forwards_idx = idx+1 if idx else null (there is no newer)
    # end_idx = -1 
    # back = idx-1 if idx else Null (there are no older)
    # begining = 0
    history = session["annotation_history"]
    current_id = post["paragraph_id"]
    session["history_dive"] = True
    if history:
        # print(history)
        #Hvis sætningen er del af sæssions historien
        #Altså hvis det er en sætning, der har været set på før
        idx = history.index(current_id) if current_id in history else -1

        #Remember the frontend screens for errors like at the end or begining of list
        if idx != -1:
            if move == "forward":
                idx += 1
            elif move == "back":
                idx -= 1
            elif move == "end":
                pass
            elif move == "begining":
                pass
        print("HISTORY")
        print(history)
        print("INDEX")
        print(idx)
        new_paragraph_id = history[idx]
        print(new_paragraph_id)
        return(new_paragraph_id)
    else: 
        print("_"*70)
        print("THERE IS NO HISTORY")
        
def create_dataframe_template():
    return(DataFrame(columns=["paragraph_text", "document_name", "page_nr"]))
    
def get_labton_data(project_name="labton_default_project"):
    app = Flask(__name__)
    ch = ConfigHandler(project_name=project_name)
    ch.load_project_config()
    app.config.update(ch.config)
    with app.app_context():
        with DatabaseHandler() as DH:
            return(DH.load_annotations_to_pandas())